# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Validator class tests
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2017, 2019, 2020 Lele Gaifax
#

import pytest

import rapidjson as rj


def test_invalid_schema():
    pytest.raises(rj.JSONDecodeError, rj.Validator, '')
    pytest.raises(rj.JSONDecodeError, rj.Validator, '"')


def test_invalid_json():
    validate = rj.Validator('""')
    pytest.raises(rj.JSONDecodeError, validate, '')
    pytest.raises(rj.JSONDecodeError, validate, '"')


@pytest.mark.parametrize('schema,json', (
    ('{ "type": ["number", "string"] }', '42'),
    ('{ "type": ["number", "string"] }', '"Life, the universe, and everything"'),
))
def test_valid(schema, json):
    validate = rj.Validator(schema)
    validate(json)


@pytest.mark.parametrize('schema,json,details', (
    ('{ "type": ["number", "string"] }',
     '["Life", "the universe", "and everything"]',
     ('type', '#', '#'),
    ),
))
def test_invalid(schema, json, details):
    validate = rj.Validator(schema)
    with pytest.raises(ValueError) as error:
        validate(json)
    assert error.value.args == details


# See: https://spacetelescope.github.io/understanding-json-schema/reference/object.html#pattern-properties
@pytest.mark.parametrize('schema', [
    rj.dumps({
        "type": "object",
        "patternProperties": {
            "^S_": { "type": "string" },
            "^I_": { "type": "integer" }
        },
        "additionalProperties": False
    }),
])
@pytest.mark.parametrize('json', [
     '{"I_0": 23}',
     '{"S_1": "the quick brown fox jumps over the lazy dog"}',
     pytest.param('{"I_2": "A string"}', marks=pytest.mark.xfail),
     pytest.param('{"keyword": "value"}', marks=pytest.mark.xfail),
])
def test_additional_and_pattern_properties_valid(schema, json):
    validate = rj.Validator(schema)
    validate(json)



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Tracemalloc-based leaks tests
# :Created:   dom 10 feb 2019 13:47:32 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2019 Lele Gaifax
#

import io
import datetime
import gc

import pytest
import rapidjson as rj

tracemalloc = pytest.importorskip("tracemalloc")


def object_hook(td):
    if '__td__' in td:
        return datetime.timedelta(td['__td__'])
    else:
        return td


def default(obj):
    if isinstance(obj, datetime.timedelta):
        return {"__td__": obj.total_seconds()}
    else:
        return obj


def test_object_hook_and_default():
    tracemalloc.start()

    data = []
    for i in range(1, 100):
        data.append({"name": "a%d" % i, "timestamp": datetime.timedelta(seconds=i)})

    snapshot1 = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(True, __file__),))

    for _ in range(1000):
        a = rj.dumps(data, default=default)
        rj.loads(a, object_hook=object_hook)

    gc.collect()

    snapshot2 = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(True, __file__),))

    top_stats = snapshot2.compare_to(snapshot1, 'lineno')
    tracemalloc.stop()

    for stat in top_stats[:10]:
        assert stat.count_diff < 3


def test_load():
    tracemalloc.start()

    snapshot1 = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(True, __file__),))

    for _ in range(10):
        dct = '{' + ','.join('"foo%d":"bar%d"' % (i, i) for i in range(100)) + '}'
        content = io.StringIO('[' + ','.join(dct for _ in range(100)) + ']')
        rj.load(content, chunk_size=50)

    del content
    del _
    gc.collect()

    snapshot2 = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(True, __file__),))

    top_stats = snapshot2.compare_to(snapshot1, 'lineno')
    tracemalloc.stop()

    for stat in top_stats[:10]:
        assert stat.count_diff < 3


def test_failed_validation():
    tracemalloc.start()

    schema = """{
        "$schema": "http://json-schema.org/draft-04/schema#",
        "required": ["id", "name"],
        "type": "object",
        "properties": {
            "id": {"type": "integer"},
            "name": {"type": "string"}
        }
    }""".encode("utf-8")

    obj = """{
        "id": 50
    }""".encode("utf-8")

    validate = rj.Validator(schema)

    snapshot1 = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(True, __file__),))

    # start the test
    for j in range(1000):
        try:
            validate(obj)
        except rj.ValidationError:
            pass

    del j

    gc.collect()

    snapshot2 = tracemalloc.take_snapshot().filter_traces((
        tracemalloc.Filter(True, __file__),))

    top_stats = snapshot2.compare_to(snapshot1, 'lineno')
    tracemalloc.stop()

    for stat in top_stats[:10]:
        assert stat.count_diff < 3



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Tests configuration
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2016, 2017, 2018 Lele Gaifax
#

import io
import rapidjson as rj


def binary_streaming_dumps(o, **opts):
    stream = io.BytesIO()
    rj.dump(o, stream, **opts)
    return stream.getvalue().decode('utf-8')


def text_streaming_dumps(o, **opts):
    stream = io.StringIO()
    rj.dump(o, stream, **opts)
    return stream.getvalue()


def binary_streaming_encoder(o, **opts):
    stream = io.BytesIO()
    rj.Encoder(**opts)(o, stream=stream)
    return stream.getvalue().decode('utf-8')


def text_streaming_encoder(o, **opts):
    stream = io.StringIO()
    rj.Encoder(**opts)(o, stream=stream)
    return stream.getvalue()


def pytest_generate_tests(metafunc):
    if 'dumps' in metafunc.fixturenames and 'loads' in metafunc.fixturenames:
        metafunc.parametrize('dumps,loads', (
            ((rj.dumps, rj.loads),
             (lambda o,**opts: rj.Encoder(**opts)(o),
              lambda j,**opts: rj.Decoder(**opts)(j)))
        ), ids=('func[string]',
                'class[string]'))
    elif 'dumps' in metafunc.fixturenames:
        metafunc.parametrize('dumps', (
            rj.dumps,
            binary_streaming_dumps,
            text_streaming_dumps,
            lambda o,**opts: rj.Encoder(**opts)(o),
            binary_streaming_encoder,
            text_streaming_encoder,
        ), ids=('func[string]',
                'func[bytestream]',
                'func[textstream]',
                'class[string]',
                'class[binarystream]',
                'class[textstream]'))
    elif 'loads' in metafunc.fixturenames:
        metafunc.parametrize('loads', (
            rj.loads,
            lambda j,**opts: rj.load(io.BytesIO(j.encode('utf-8')
                                           if isinstance(j, str) else j), **opts),
            lambda j,**opts: rj.load(io.StringIO(j), **opts),
            lambda j,**opts: rj.Decoder(**opts)(j),
            lambda j,**opts: rj.Decoder(**opts)(io.BytesIO(j.encode('utf-8')
                                                      if isinstance(j, str) else j)),
            lambda j,**opts: rj.Decoder(**opts)(io.StringIO(j)),
        ), ids=('func[string]',
                'func[bytestream]',
                'func[textstream]',
                'class[string]',
                'class[bytestream]',
                'class[textstream]'))



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Streaming API related tests
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2017, 2018, 2020 Lele Gaifax
#

import io
import sys
import tempfile

import pytest

import rapidjson as rj


@pytest.mark.parametrize('cs', (-1, 0, sys.maxsize*10, 1.23, 'foo'))
def test_invalid_chunk_size(cs):
    s = io.StringIO('"foo"')
    with pytest.raises((ValueError, TypeError)):
        rj.load(s, chunk_size=cs)


class ChunkedStream(io.StringIO):
    def __init__(self):
        super().__init__()
        self.chunks = []

    def write(self, s):
        super().write(s)
        self.chunks.append(s)


def test_chunked_stream():
    stream = ChunkedStream()
    rj.dump('1234567890', stream)
    assert len(stream.chunks) == 1

    stream = ChunkedStream()
    rj.dump('1234567890', stream, chunk_size=4)
    assert len(stream.chunks) == 3
    assert stream.chunks == ['"123', '4567', '890"']

    stream = ChunkedStream()
    rj.dump('~𓆙~', stream, ensure_ascii=False, chunk_size=4)
    assert len(stream.chunks) == 3
    assert stream.chunks == ['"~', '𓆙', '~"']

    stream = ChunkedStream()
    rj.dump('~𓆙~', stream, chunk_size=4)
    assert len(stream.chunks) == 4
    assert stream.chunks == ['"~\\u', 'D80C', '\\uDD', '99~"']


class CattyError(RuntimeError):
    pass


class CattyStream(io.StringIO):
    def read(self, *args, **kwargs):
        raise CattyError('No real reason')

    def write(self, *args, **kwargs):
        raise CattyError('No real reason')


def test_underlying_stream_read_error():
    stream = CattyStream()
    with pytest.raises(CattyError):
        rj.load(stream)


def test_underlying_stream_write_error():
    stream = CattyStream()
    with pytest.raises(CattyError):
        rj.dump('1234567890', stream)


def test_file_object():
    for stream in tempfile.TemporaryFile(), tempfile.TemporaryFile('w+', encoding='utf-8'):
        with stream:
            datum = ['1234567890', 1234, 3.14, '~𓆙~']
            rj.dump(datum, stream)
            stream.seek(0)
            assert rj.load(stream) == datum



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Test dict subclasses
# :Created:   dom 17 mar 2019 12:38:24 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2019, 2021 Lele Gaifax
#

from collections import OrderedDict
from rapidjson import Decoder


class OrderedDecoder(Decoder):
    def start_object(self):
        return OrderedDict()


def test_ordered_dict():
    od = OrderedDecoder()
    result = od('{"foo": "bar", "bar": "foo"}')
    assert isinstance(result, OrderedDict)
    assert list(result.keys()) == ["foo", "bar"]


class ObjectsAsKeyValuePairsDecoder(Decoder):
    def start_object(self):
        return []


def test_objects_as_key_value_pairs():
    kvp = ObjectsAsKeyValuePairsDecoder()
    result = kvp('{"a": 1, "b": {"b1": 1, "b2": 2}}')
    assert result == [('a', 1), ('b', [('b1', 1), ('b2', 2)])]


class JoinDuplicatedKeysDecoder(ObjectsAsKeyValuePairsDecoder):
    def end_object(self, ordered_pairs):
        # Adapted from https://stackoverflow.com/a/38307621
        d = {}
        for k, v in ordered_pairs:
            if k in d:
                if type(d[k]) == list:
                    d[k].append(v)
                else:
                    newlist = []
                    newlist.append(d[k])
                    newlist.append(v)
                    d[k] = newlist
            else:
                d[k] = v
        return d


def test_join_duplicated_keys():
    jdk = JoinDuplicatedKeysDecoder()
    result = jdk('{"a": 1, "b": {"b1": 1, "b2": 2}, "b": {"b1": 3, "b2": 2,"b4": 8}}')
    assert result == {'a': 1, 'b': [{'b1': 1, 'b2': 2}, {'b1': 3, 'b2': 2, 'b4': 8}]}



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- reST tables writer tool
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2017, 2018, 2020, 2021, 2022 Lele Gaifax
#

# Assume we did::
#
#  pytest --benchmark-json=comparison.json --compare-other-engines benchmarks/
#
# Read back that JSON outcome and produce three reST tables summarizing the
# comparison.
#

from collections import defaultdict, namedtuple
import sys

import rapidjson


Benchmark = namedtuple('Benchmark', 'group, name, contender')


class Comparison:
    COMPARE_FIELD = 'min'  # one of min, max, mean, median...

    def __init__(self, contenders, benchmarks):
        self.contenders = contenders
        self.timings = {}
        self.load_data(benchmarks)

    def load_data(self, benchmarks):
        for bm in benchmarks:
            group = bm['group']
            desc = bm['name'].split('[', 1)[1][:-1]
            contender, name = desc.split('-')
            if contender in self.contenders:
                bmark = Benchmark(group, name, contender)
                self.timings[bmark] = bm['stats'][self.COMPARE_FIELD]

    def normalize_and_summarize(self, group):
        table = defaultdict(dict)
        for bmark, timing in self.timings.items():
            if bmark.group == group:
                table[bmark.name][bmark.contender] = timing

        normalized = {}
        sums = defaultdict(float)
        by = self.contenders[0]
        for name, values in table.items():
            for c in self.contenders:
                sums[c] += values[c]
            normalized[name] = [values[c] / values[by] for c in self.contenders]

        normalized_sum = [sums[c] / sums[by] for c in self.contenders]

        return normalized, normalized_sum

    def tabulate(self, group, headers):
        normalized, sum = self.normalize_and_summarize(group)

        nc = len(self.contenders)
        widths = [max((len(n)+4 for n in normalized))] + \
                 [max((len(h) for h in headers))] * nc

        cells = ['-' * (width+2) for width in widths]
        seprow = '+' + '+'.join(cells) + '+'

        cells = ['=' * (width+2) for width in widths]
        headseprow = '+' + '+'.join(cells) + '+'

        cells = [' {:^%d} ' % width for width in widths]
        rowfmt = '|' + '|'.join(cells) + '|'

        print(seprow)
        print(rowfmt.format(*([group] + [headers[i]
                                         for i, c in enumerate(self.contenders)])))
        print(headseprow)

        def printrow(name, row):
            best = min(row)
            fvalues = [('**%.2f**' if v==best else '%.2f') % v for v in row]
            print(rowfmt.format(*([name] + fvalues)))
            print(seprow)

        for n in sorted(normalized):
            printrow(n, normalized[n])

        printrow('overall', sum)


def dumps_and_loads(benchmarks):
    contenders = ('rapidjson_f',
                  'rapidjson_c',
                  'rapidjson_nn_f',
                  'rapidjson_nn_c',
                  'simdjson',
                  'orjson',
                  'ujson',
                  'simplejson',
                  'stdlib json')
    s_headers = (r'``dumps()``\ [1]_',
                 r'``Encoder()``\ [2]_',
                 r'``dumps(n)``\ [3]_',
                 r'``Encoder(n)``\ [4]_',
                 r'simdjson\ [5]_',
                 r'orjson\ [6]_',
                 r'ujson\ [7]_',
                 r'simplejson\ [8]_',
                 r'stdlib\ [9]_')
    d_headers = (r'``loads()``\ [10]_',
                 r'``Decoder()``\ [11]_',
                 r'``loads(n)``\ [12]_',
                 r'``Decoder(n)``\ [13]_',
                 r'simdjson',
                 r'orjson',
                 r'ujson',
                 r'simplejson',
                 r'stdlib')

    assert len(contenders) == len(s_headers) == len(d_headers)

    footnotes = [
        '.. [1] ``rapidjson.dumps()``',
        '.. [2] ``rapidjson.Encoder()``',
        '.. [3] ``rapidjson.dumps(number_mode=NM_NATIVE)``',
        '.. [4] ``rapidjson.Encoder(number_mode=NM_NATIVE)``',
        '.. [5] `simdjson 5.0.1 <https://pypi.org/project/pysimdjson/5.0.1/>`__',
        '.. [6] `orjson 3.7.6 <https://pypi.org/project/orjson/3.7.6/>`__',
        '.. [7] `ujson 5.4.0 <https://pypi.org/project/ujson/5.4.0/>`__',
        '.. [8] `simplejson 3.17.6 <https://pypi.org/pypi/simplejson/3.17.6>`__',
        '.. [9] Python %d.%d.%d standard library ``json``' % sys.version_info[:3],
        '.. [10] ``rapidjson.loads()``',
        '.. [11] ``rapidjson.Decoder()``',
        '.. [12] ``rapidjson.loads(number_mode=NM_NATIVE)``',
        '.. [13] ``rapidjson.Decoder(number_mode=NM_NATIVE)``']

    comparison = Comparison(contenders, benchmarks)

    print()
    print('Serialization')
    print('~~~~~~~~~~~~~')
    print()
    comparison.tabulate('serialize', s_headers)

    print()
    print('Deserialization')
    print('~~~~~~~~~~~~~~~')
    print()
    comparison.tabulate('deserialize', d_headers)
    return footnotes


def ascii_vs_utf8(benchmarks):
    engines = (('rapidjson', 'rj'),
               ('ujson', 'uj'),
               ('simplejson', 'sj'),
               ('stdlib json', 'json'))
    contenders = ['%s %s' % (e[0], k) for e in engines for k in ('ascii', 'utf8')]
    i = 14
    s_headers = []
    footnotes = []
    for e in engines:
        i += 1
        s_headers.append(r'``%s ascii``\ [%d]_' % (e[1], i))
        footnotes.append('.. [%d] ``%s.dumps(ensure_ascii=True)``' % (i, e[0]))
        i += 1
        s_headers.append(r'``%s utf8``\ [%d]_' % (e[1], i))
        footnotes.append('.. [%d] ``%s.dumps(ensure_ascii=False)``' % (i, e[0]))

    comparison = Comparison(contenders, benchmarks)

    print()
    print('ASCII vs UTF-8 Serialization')
    print('~~~~~~~~~~~~~~~~~~~~~~~~~~~~')
    print()
    comparison.tabulate('serialize', s_headers)
    return footnotes


def main():
    import sys

    if len(sys.argv) > 1:
        fname = sys.argv[1]
    else:
        fname = 'comparison.json'

    with open(fname) as f:
        benchmarks = rapidjson.loads(f.read())['benchmarks']

    footnotes = []

    footnotes.extend(dumps_and_loads(benchmarks))
    footnotes.extend(ascii_vs_utf8(benchmarks))

    print()
    for fn in footnotes:
        print(fn)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Tests on RawJSON
# :Author:    Silvio Tomatis <silviot@gmail.com>
# :License:   MIT License
# :Copyright: © 2018 Silvio Tomatis
# :Copyright: © 2018, 2020 Lele Gaifax
#

import pytest

import rapidjson as rj


def test_instantiation_positional():
    value = 'a string'
    raw = rj.RawJSON(value)
    assert raw.value == value


def test_instantiation_named():
    value = 'a string'
    raw = rj.RawJSON(value=value)
    assert raw.value == value


def test_only_bytes_allowed():
    with pytest.raises(TypeError):
        rj.RawJSON(b'A bytes string')
    with pytest.raises(TypeError):
        rj.RawJSON({})
    with pytest.raises(TypeError):
        rj.RawJSON(None)


def test_mix_preserialized():
    assert rj.dumps({'a': rj.RawJSON('{1 : 2}')}) == '{"a":{1 : 2}}'



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Circular references tests
# :Author:    John Anderson <sontek@gmail.com>
# :License:   MIT License
# :Copyright: © 2015 John Anderson
# :Copyright: © 2017, 2018, 2020 Lele Gaifax
#

import sys

import pytest


if sys.version_info >= (3, 5):
    expected_exception = RecursionError
else:
    expected_exception = RuntimeError


def test_circular_dict(dumps):
    dct = {}
    dct['a'] = dct

    with pytest.raises(expected_exception):
        dumps(dct)


def test_circular_list(dumps):
    lst = []
    lst.append(lst)

    with pytest.raises(expected_exception):
        dumps(lst)


def test_circular_composite(dumps):
    dct2 = {}
    dct2['a'] = []
    dct2['a'].append(dct2)

    with pytest.raises(expected_exception):
        dumps(dct2)


def test_max_recursion_depth(dumps):
    import sys

    root = child = {}
    for i in range(500):
        nephew = {'value': i}
        child['child'] = nephew
        child = nephew

    dumps(root)

    rl = sys.getrecursionlimit()

    if hasattr(sys, "pyston_version_info"):
        # In Pyston setrecursionlimit() sets a lower bound on the number of frames,
        # not an exact count. So set the limit lower:
        sys.setrecursionlimit(100)
    else:
        # Otherwise set it to the exact value:
        sys.setrecursionlimit(500)

    try:
        with pytest.raises(expected_exception):
            dumps(root)
    finally:
        sys.setrecursionlimit(rl)



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Benchmarks specific pytest configuration
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2016, 2017, 2018, 2020 Lele Gaifax
#

from collections import namedtuple
from functools import partial
from operator import attrgetter

Contender = namedtuple('Contender', 'name,dumps,loads')


# See https://github.com/ionelmc/pytest-benchmark/issues/48

def pytest_benchmark_group_stats(config, benchmarks, group_by):
    result = {}
    for bench in benchmarks:
        engine, data_kind = bench['param'].split('-')
        group = result.setdefault("%s: %s" % (data_kind, bench['group']), [])
        group.append(bench)
    return sorted(result.items())


# Remove parametrization data from JSON output to keep it to a reasonable size
# See https://github.com/ionelmc/pytest-benchmark/issues/96

def pytest_benchmark_update_json(config, benchmarks, output_json):
    for benchmark in output_json['benchmarks']:
        if 'data' in benchmark['params']:
            benchmark['params'].pop('data')
        if 'data' in benchmark['stats']:
            benchmark['stats'].pop('data')


def pytest_addoption(parser):
    parser.addoption('--compare-other-engines', action='store_true',
                     help='compare against other JSON engines')


contenders = []

import rapidjson as rj

contenders.append(Contender('rapidjson_f',
                            rj.dumps,
                            rj.loads))
contenders.append(Contender('rapidjson_c',
                            rj.Encoder(),
                            rj.Decoder()))
contenders.append(Contender('rapidjson_nn_f',
                            partial(rj.dumps, number_mode=rj.NM_NATIVE),
                            partial(rj.loads, number_mode=rj.NM_NATIVE)))
contenders.append(Contender('rapidjson_nn_c',
                            rj.Encoder(number_mode=rj.NM_NATIVE),
                            rj.Decoder(number_mode=rj.NM_NATIVE)))

numbers_contenders = [
    Contender('Wide numbers', rj.dumps, rj.loads),
    Contender('Native numbers',
              partial(rj.dumps, number_mode=rj.NM_NATIVE),
              partial(rj.loads, number_mode=rj.NM_NATIVE)),
]

string_contenders = [
    Contender('rapidjson utf8',
              partial(rj.dumps, ensure_ascii=False),
              rj.loads),
    Contender('rapidjson ascii',
              partial(rj.dumps, ensure_ascii=True),
              rj.loads),
]

try:
    import simplejson
except ImportError:
    pass
else:
    contenders.append(Contender('simplejson',
                                simplejson.dumps,
                                simplejson.loads))
    string_contenders.extend([
        Contender('simplejson utf8',
                  partial(simplejson.dumps, ensure_ascii=False),
                  simplejson.loads),
        Contender('simplejson ascii',
                  partial(simplejson.dumps, ensure_ascii=True),
                  simplejson.loads),
    ])

try:
    import json
except ImportError:
    pass
else:
    contenders.append(Contender('stdlib json',
                                json.dumps,
                                json.loads))
    string_contenders.extend([
        Contender('stdlib json utf8',
                  partial(json.dumps, ensure_ascii=False),
                  json.loads),
        Contender('stdlib json ascii',
                  partial(json.dumps, ensure_ascii=True),
                  json.loads),
    ])

try:
    import ujson
except ImportError:
    pass
else:
    contenders.append(Contender('ujson',
                                ujson.dumps,
                                ujson.loads))
    string_contenders.extend([
        Contender('ujson utf8',
                  partial(ujson.dumps, ensure_ascii=False),
                  ujson.loads),
        Contender('ujson ascii',
                  partial(ujson.dumps, ensure_ascii=True),
                  ujson.loads),
    ])

try:
    import simdjson
except ImportError:
    pass
else:
    contenders.append(Contender('simdjson',
                                simdjson.dumps,
                                simdjson.loads))

try:
    import orjson
except ImportError:
    pass
else:
    contenders.append(Contender('orjson',
                                orjson.dumps,
                                orjson.loads))


def pytest_generate_tests(metafunc):
    if 'contender' in metafunc.fixturenames:
        if metafunc.config.option.compare_other_engines:
            metafunc.parametrize('contender', contenders, ids=attrgetter('name'))
        else:
            metafunc.parametrize('contender', contenders[:1], ids=attrgetter('name'))

    if 'datetimes_loads_contender' in metafunc.fixturenames:
        metafunc.parametrize('datetimes_loads_contender',
                             [rj.loads,
                              partial(rj.loads, datetime_mode=rj.DM_ISO8601)],
                             ids=['Ignore datetimes', 'Parse datetimes'])

    if 'numbers_contender' in metafunc.fixturenames:
        metafunc.parametrize('numbers_contender', numbers_contenders, ids=attrgetter('name'))

    if 'string_contender' in metafunc.fixturenames:
        if metafunc.config.option.compare_other_engines:
            metafunc.parametrize('string_contender', string_contenders, ids=attrgetter('name'))
        else:
            metafunc.parametrize('string_contender',
                                 [c for c in string_contenders if 'rapidjson' in c[0]],
                                 ids=attrgetter('name'))



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Tests on floats
# :Author:    John Anderson <sontek@gmail.com>
# :License:   MIT License
# :Copyright: © 2015 John Anderson
# :Copyright: © 2016, 2017, 2020 Lele Gaifax
#

from decimal import Decimal
import math

import pytest

import rapidjson as rj


def test_infinity_f():
    inf = float("inf")

    dumped = rj.dumps(inf)
    loaded = rj.loads(dumped)
    assert loaded == inf

    dumped = rj.dumps(inf, allow_nan=True)
    loaded = rj.loads(dumped, allow_nan=True)
    assert loaded == inf

    with pytest.raises(ValueError):
        rj.dumps(inf, number_mode=None)

    with pytest.raises(ValueError):
        rj.dumps(inf, allow_nan=False)

    d = Decimal(inf)
    assert d.is_infinite()

    with pytest.raises(ValueError):
        rj.dumps(d, number_mode=rj.NM_DECIMAL)

    dumped = rj.dumps(d, number_mode=rj.NM_DECIMAL, allow_nan=True)
    loaded = rj.loads(dumped, number_mode=rj.NM_DECIMAL|rj.NM_NAN)
    assert loaded == inf
    assert loaded.is_infinite()


def test_infinity_c():
    inf = float("inf")

    dumped = rj.Encoder()(inf)
    loaded = rj.Decoder()(dumped)
    assert loaded == inf

    with pytest.raises(ValueError):
        rj.Encoder(number_mode=None)(inf)

    d = Decimal(inf)
    assert d.is_infinite()

    with pytest.raises(ValueError):
        rj.Encoder(number_mode=rj.NM_DECIMAL)(d)

    dumped = rj.Encoder(number_mode=rj.NM_DECIMAL|rj.NM_NAN)(d)
    loaded = rj.Decoder(number_mode=rj.NM_DECIMAL|rj.NM_NAN)(dumped)
    assert loaded == inf
    assert loaded.is_infinite()


def test_negative_infinity_f():
    inf = float("-infinity")
    dumped = rj.dumps(inf)
    loaded = rj.loads(dumped)
    assert loaded == inf

    dumped = rj.dumps(inf, allow_nan=True)
    loaded = rj.loads(dumped, allow_nan=True)
    assert loaded == inf

    with pytest.raises(ValueError):
        rj.dumps(inf, number_mode=None)

    with pytest.raises(ValueError):
        rj.dumps(inf, allow_nan=False)

    d = Decimal(inf)
    assert d.is_infinite()

    with pytest.raises(ValueError):
        rj.dumps(d, number_mode=rj.NM_DECIMAL)

    dumped = rj.dumps(d, number_mode=rj.NM_DECIMAL|rj.NM_NAN)
    loaded = rj.loads(dumped, number_mode=rj.NM_DECIMAL, allow_nan=True)
    assert loaded == inf
    assert loaded.is_infinite()


def test_negative_infinity_c():
    inf = float("-infinity")
    dumped = rj.Encoder()(inf)
    loaded = rj.Decoder()(dumped)
    assert loaded == inf

    with pytest.raises(ValueError):
        rj.Encoder(number_mode=None)(inf)

    d = Decimal(inf)
    assert d.is_infinite()

    with pytest.raises(ValueError):
        rj.Encoder(number_mode=rj.NM_DECIMAL)(d)

    dumped = rj.Encoder(number_mode=rj.NM_DECIMAL|rj.NM_NAN)(d)
    loaded = rj.Decoder(number_mode=rj.NM_DECIMAL|rj.NM_NAN)(dumped)
    assert loaded == inf
    assert loaded.is_infinite()


def test_nan_f():
    nan = float("nan")
    dumped = rj.dumps(nan)
    loaded = rj.loads(dumped)

    assert math.isnan(nan)
    assert math.isnan(loaded)

    with pytest.raises(ValueError):
        rj.dumps(nan, number_mode=None)

    with pytest.raises(ValueError):
        rj.dumps(nan, allow_nan=False)

    d = Decimal(nan)
    assert d.is_nan()

    with pytest.raises(ValueError):
        rj.dumps(d, number_mode=rj.NM_DECIMAL)

    dumped = rj.dumps(d, number_mode=rj.NM_DECIMAL|rj.NM_NAN)
    loaded = rj.loads(dumped, number_mode=rj.NM_DECIMAL|rj.NM_NAN)
    assert loaded.is_nan()


def test_nan_c():
    nan = float("nan")
    dumped = rj.Encoder()(nan)
    loaded = rj.Decoder()(dumped)

    assert math.isnan(nan)
    assert math.isnan(loaded)

    with pytest.raises(ValueError):
        rj.Encoder(number_mode=None)(nan)

    d = Decimal(nan)
    assert d.is_nan()

    with pytest.raises(ValueError):
        rj.Encoder(number_mode=rj.NM_DECIMAL)(d)

    dumped = rj.Encoder(number_mode=rj.NM_DECIMAL|rj.NM_NAN)(d)
    loaded = rj.Decoder(number_mode=rj.NM_DECIMAL|rj.NM_NAN)(dumped)
    assert loaded.is_nan()



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Basic tests
# :Author:    John Anderson <sontek@gmail.com>
# :License:   MIT License
# :Copyright: © 2015 John Anderson
# :Copyright: © 2016, 2017, 2018, 2019, 2020, 2021 Lele Gaifax
#

import random
import sys

import pytest

import rapidjson as rj


@pytest.mark.parametrize(
    'value', (
        'A', 'cruel\x00world',
        1, -1,
        2.3, -36.973846435546875, 1514893636.276703,
        {'foo': 'bar', '\x00': 'issue57', 'issue57': '\x00'},
        [1, 2, 'a', 1.2, {'foo': 'bar'},],
        sys.maxsize, sys.maxsize**2, 10**1000,
))
def test_base_values(value, dumps, loads):
    dumped = dumps(value)
    loaded = loads(dumped)
    assert loaded == value and type(loaded) is type(value)


def test_float(dumps):
    value = 0.1 + 0.2
    dumped = dumps(value)
    assert dumped == '0.30000000000000004'


def test_tuple(dumps):
    obj = [1, 2, 'a', 1.2, {'foo': 'bar'},]
    assert dumps(obj) == dumps(tuple(obj))


def test_bytes_value(dumps):
    value = b'cruel\x00world'
    dumped = dumps(value)
    assert dumped == r'"cruel\u0000world"'
    dumped = dumps(bytearray(value))
    assert dumped == r'"cruel\u0000world"'


def test_larger_structure(dumps, loads):
    value = {
        'words': """
            Lorem ipsum dolor sit amet, consectetur adipiscing
            elit. Mauris adipiscing adipiscing placerat.
            Vestibulum augue augue,
            pellentesque quis sollicitudin id, adipiscing.
            """,
        'list': list(range(200)),
        'dict': dict((str(i),'a') for i in range(200)),
        'int': 100100100,
        'float': 100999.123456
    }

    dumped = dumps(value)
    loaded = loads(dumped)
    assert loaded == value


def test_object_hook():
    def as_complex(dct):
        if '__complex__' in dct:
            return complex(dct['real'], dct['imag'])

        return dct

    result = rj.loads(
        '{"__complex__": true, "real": 1, "imag": 2}',
        object_hook=as_complex
    )

    assert result == (1+2j)


def test_end_object():
    class ComplexDecoder(rj.Decoder):
        def end_object(self, dct):
            if '__complex__' in dct:
                return complex(dct['real'], dct['imag'])

            return dct

    loads = ComplexDecoder()
    result = loads('{"__complex__": true, "real": 1, "imag": 2}')
    assert result == (1+2j)


def test_dumps_default():
    def encode_complex(obj):
        if isinstance(obj, complex):
            return [obj.real, obj.imag]
        raise TypeError(repr(obj) + " is not JSON serializable")

    result = rj.dumps(2 + 1j, default=encode_complex)
    assert result == '[2.0,1.0]'


def test_encoder_default():
    class ComplexEncoder(rj.Encoder):
        def default(self, obj):
            if isinstance(obj, complex):
                return [obj.real, obj.imag]
            raise TypeError(repr(obj) + " is not JSON serializable")

    dumps = ComplexEncoder()
    result = dumps(2 + 1j)
    assert result == '[2.0,1.0]'


def test_doubles(dumps, loads):
    for _ in range(100000):
        d = sys.maxsize * random.random()
        dumped = dumps(d)
        loaded = loads(dumped)
        assert loaded == d


def test_unicode(dumps, loads):
   arabic='بينهم ان يكون مسلما رشيدا عاقلا ًوابنا شرعيا لابوين عمانيين'
   chinese='本站所提供的資料和服務都不收費，因此網站所需要的資金全來自廣告及捐款。若您願意捐款補助'
   for text in (arabic, chinese):
       dumped = dumps(text)
       loaded = loads(dumped)
       assert text == loaded


def test_serialize_sets_dumps():
    def default_iterable(obj):
        if isinstance(obj, set):
            return list(obj)
        raise TypeError(repr(obj) + " is not JSON serializable")

    rj.dumps([set()], default=default_iterable)

    with pytest.raises(TypeError):
        rj.dumps([set()])


def test_serialize_sets_encoder():
    class SetsEncoder(rj.Encoder):
        def default(self, obj):
            if isinstance(obj, set):
                return list(obj)
            raise TypeError(repr(obj) + " is not JSON serializable")

    dumps = SetsEncoder()
    dumps([set()])

    with pytest.raises(TypeError):
        rj.Encoder()([set()])


def test_constants(dumps, loads):
    for c in [None, True, False]:
        assert loads(dumps(c)) is c
        assert loads(dumps([c]))[0] is c
        assert loads(dumps({'a': c}))['a'] is c


def test_iterables(dumps):
    assert dumps(iter("abc")) == '["a","b","c"]'

    def gen():
        yield 1
        yield 2
        yield 3

    assert dumps(gen()) == '[1,2,3]'


def test_decode_error(loads):
    pytest.raises(rj.JSONDecodeError, loads, '{')


def test_shared_keys(loads):
    res = loads('[{"key": "value1"}, {"key": "value2"}]')
    key1, = res[0].keys()
    key2, = res[1].keys()
    assert key1 is key2


# TODO: Figure out what we want to do here
bad_tests = """
def test_true_false():
    dumped1 = sorted(rj.dumps({True: False, False: True}))
    dumped2 = sorted(rj.dumps({
        2: 3.0,
        4.0: long_type(5),
        False: 1,
        long_type(6): True,
        "7": 0
    }))

    assert dumped1 == '{"false": true, "true": false}'
    expected = '{"2": 3.0, "4.0": 5, "6": true, "7": 0, "false": 1}'

    assert dumped2 == expected"""



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Test on stdlib enums
# :Created:   mer 27 mar 2019 08:13:23 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2019, 2020 Lele Gaifax
#

import enum

import rapidjson as rj


def test_intenums_as_ints():
    class IE(enum.IntEnum):
        val = 123
        bigval = 123123123123123123123123

    assert rj.dumps([IE.val, IE.bigval]) == "[123,123123123123123123123123]"



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Unicode tests
# :Author:    John Anderson <sontek@gmail.com>
# :License:   MIT License
# :Copyright: © 2015 John Anderson
# :Copyright: © 2016, 2017, 2018, 2020 Lele Gaifax
#

import json

import pytest

import rapidjson


@pytest.mark.parametrize('u', [
    '\N{GREEK SMALL LETTER ALPHA}\N{GREEK CAPITAL LETTER OMEGA}',
    '\U0010ffff',
    'asdf \U0010ffff \U0001ffff qwert \uffff \u10ff \u00ff \u0080 \u7fff \b\n\r',
])
def test_unicode(u, dumps, loads):
    s = u.encode('utf-8')
    ju = dumps(u)
    js = dumps(s)
    assert ju == js
    assert ju.lower() == json.dumps(u).lower()
    assert dumps(u, ensure_ascii=False) == json.dumps(u, ensure_ascii=False)


@pytest.mark.parametrize('o', [
    "\ud80d",
    {"foo": "\ud80d"},
    {"\ud80d": "foo"},
])
def test_dump_surrogate(o, dumps):
    with pytest.raises(UnicodeEncodeError, match="surrogates not allowed"):
        dumps(o)


@pytest.mark.parametrize('j', [
    '"\\ud80d"',
    '{"foo": "\\ud80d"}',
    '{"\\ud80d": "foo"}',
])
def test_load_surrogate(j, loads):
    with pytest.raises(ValueError, match="surrogate pair in string is invalid"):
        loads(j)


@pytest.mark.parametrize('j', [
    '"\\udc00"',
    '"\\udfff"',
])
def test_unicode_decode_error(j, loads):
    with pytest.raises(rapidjson.JSONDecodeError,
                       match="The surrogate pair in string is invalid."):
        loads(j)


def test_non_utf8_bytes(dumps):
    value = b'\xff\xf0'
    with pytest.raises(UnicodeDecodeError, match="'utf-8' codec can't decode byte"):
        dumps(value)



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Compliance tests
# :Author:    John Anderson <sontek@gmail.com>
# :License:   MIT License
# :Copyright: © 2015 John Anderson
# :Copyright: © 2017, 2020 Lele Gaifax
#

# from http://json.org/JSON_checker/test/pass3.json
JSON = r'''
{
    "JSON Test Pattern pass3": {
        "The outermost value": "must be an object or array.",
        "In this test": "It is an object."
    }
}
'''

def test_parse(dumps, loads):
    # test in/out equivalence and parsing
    res = loads(JSON)
    out = dumps(res)
    assert res == loads(out)



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Performance tests
# :Author:    John Anderson <sontek@gmail.com>
# :License:   MIT License
# :Copyright: © 2015 John Anderson
# :Copyright: © 2015, 2016, 2017, 2018 Lele Gaifax
#

import datetime
import pathlib
import random
import sys

import pytest


benchmark = pytest.importorskip('pytest_benchmark')


composite_object = {
    'words': """
        Lorem ipsum dolor sit amet, consectetur adipiscing
        elit. Mauris adipiscing adipiscing placerat.
        Vestibulum augue augue,
        pellentesque quis sollicitudin id, adipiscing.
        """,
    'list': list(range(200)),
    'dict': dict((str(i),'a') for i in range(200)),
    'int': 100100100,
    'float': 100999.123456
}

user = {
    "userId": 3381293,
    "age": 213,
    "username": "johndoe",
    "fullname": u"John Doe the Second",
    "isAuthorized": True,
    "liked": 31231.31231202,
    "approval": 31.1471,
    "jobs": [ 1, 2 ],
    "currJob": None
}

friends = [ user, user, user, user, user, user, user, user ]

doubles = []
numbers = []
unicode_strings = []
strings = []
booleans = []
datetimes = []
list_dicts = []
dict_lists = {}
russian_string = "привет, мир!" * 1000
italian_string = "Ciao, mondo!" * 1000

complex_object = [
    [user, friends],  [user, friends],  [user, friends],
    [user, friends],  [user, friends],  [user, friends]
]

for _ in range(256):
    doubles.append(sys.maxsize * random.random())
    numbers.append(sys.maxsize * random.random())
    numbers.append(random.randint(0, sys.maxsize))
    unicode_strings.append("نظام الحكم سلطاني وراثي في الذكور من ذرية السيد تركي بن سعيد بن سلطان ويشترط فيمن يختار لولاية الحكم من بينهم ان يكون مسلما رشيدا عاقلا ًوابنا شرعيا لابوين عمانيين ")
    strings.append("A pretty long string which is in a list")
    booleans.append(True)
    datetimes.append([datetime.date(random.randint(1900, 2000),
                                    random.randint(1, 12),
                                    random.randint(1, 28)),
                      datetime.time(random.randint(0,23),
                                    random.randint(0, 59),
                                    random.randint(0, 59)),
                      datetime.datetime.now()])

for _ in range(100):
    arrays = []
    list_dicts.append({str(random.random()*20): int(random.random()*1000000)})

    for _ in range(100):
        arrays.append({str(random.random() * 20): int(random.random()*1000000)})
        dict_lists[str(random.random() * 20)] = arrays


datasets = [('composite object', composite_object),
            ('256 doubles array', doubles),
            ('256 unicode array', unicode_strings),
            ('256 ascii array', strings),
            ('256 Trues array', booleans),
            ('100 dicts array', list_dicts),
            ('100 arrays dict', dict_lists),
            ('complex object', complex_object),
]


@pytest.mark.benchmark(group='serialize')
@pytest.mark.parametrize('data', [d[1] for d in datasets], ids=[d[0] for d in datasets])
def test_dumps(contender, data, benchmark):
    benchmark(contender.dumps, data)


@pytest.mark.benchmark(group='deserialize')
@pytest.mark.parametrize('data', [d[1] for d in datasets], ids=[d[0] for d in datasets])
def test_loads(contender, data, benchmark):
    data = contender.dumps(data)
    benchmark(contender.loads, data)


# Special case 1: load datetimes as plain strings vs datetime.xxx instances

@pytest.mark.benchmark(group='deserialize')
@pytest.mark.parametrize('data', [datetimes], ids=['256x3 datetimes'])
def test_loads_datetimes(datetimes_loads_contender, data, benchmark):
    from rapidjson import dumps, DM_ISO8601
    data = dumps(data, datetime_mode=DM_ISO8601)
    benchmark(datetimes_loads_contender, data)


# Special case 2: native numbers

@pytest.mark.benchmark(group='serialize')
@pytest.mark.parametrize('data', [numbers], ids=['512 numbers array'])
def test_dumps_numbers(numbers_contender, data, benchmark):
    benchmark(numbers_contender.dumps, data)


@pytest.mark.benchmark(group='deserialize')
@pytest.mark.parametrize('data', [numbers], ids=['512 numbers array'])
def test_loads_numbers(numbers_contender, data, benchmark):
    data = numbers_contender.dumps(data)
    benchmark(numbers_contender.loads, data)


# Special case 3: ensure_ascii

@pytest.mark.benchmark(group='serialize')
@pytest.mark.parametrize('data', [italian_string], ids=['Long ASCII string'])
def test_dumps_long_plain_ascii(string_contender, data, benchmark):
    benchmark(string_contender.dumps, data)


@pytest.mark.benchmark(group='deserialize')
@pytest.mark.parametrize('data', [italian_string], ids=['Long ASCII string'])
def test_loads_long_plain_ascii(string_contender, data, benchmark):
    data = string_contender.dumps(data)
    benchmark(string_contender.loads, data)


@pytest.mark.benchmark(group='serialize')
@pytest.mark.parametrize('data', [russian_string], ids=['Long Unicode string'])
def test_dumps_long_unicode(string_contender, data, benchmark):
    benchmark(string_contender.dumps, data)


@pytest.mark.benchmark(group='deserialize')
@pytest.mark.parametrize('data', [russian_string], ids=['Long Unicode string'])
def test_loads_long_unicode(string_contender, data, benchmark):
    data = string_contender.dumps(data)
    benchmark(string_contender.loads, data)


samples = list((pathlib.Path(__file__).parent / 'json').glob('*.json'))

@pytest.mark.benchmark(group='serialize')
@pytest.mark.parametrize('sample', samples, ids=[s.name for s in samples])
def test_dumps_sample(contender, sample, benchmark):
    data = contender.loads(sample.read_text())
    benchmark(contender.dumps, data)


@pytest.mark.benchmark(group='deserialize')
@pytest.mark.parametrize('sample', samples, ids=[s.name for s in samples])
def test_loads_sample(contender, sample, benchmark):
    benchmark(contender.loads, sample.read_text())



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Packaging
# :Author:    Ken Robbins <ken@kenrobbins.com>
# :License:   MIT License
# :Copyright: © 2015 Ken Robbins
# :Copyright: © 2016, 2017, 2018, 2019, 2020, 2021, 2022 Lele Gaifax
#

import os.path
import sys

try:
    from setuptools import setup, Extension
    try:
        # This is needed for some old versions of setuptools
        import packaging.specifiers
    except ImportError:
        pass
    other_setup_options = {'python_requires': '>=3.6'}
except ImportError:
    from distutils.core import setup, Extension
    other_setup_options = {}

from distutils import sysconfig


if sys.version_info < (3, 6):
    raise NotImplementedError("Only Python 3.6+ is supported.")

ROOT_PATH = os.path.abspath(os.path.dirname(__file__))

rj_include_dir = './rapidjson/include'

for idx, arg in enumerate(sys.argv[:]):
    if arg.startswith('--rj-include-dir='):
        sys.argv.pop(idx)
        rj_include_dir = arg.split('=', 1)[1]
        break
else:
    if not os.path.isdir(os.path.join(ROOT_PATH, 'rapidjson', 'include')):
        raise RuntimeError("RapidJSON sources not found: if you cloned the git"
                           " repository, you should initialize the rapidjson submodule"
                           " as explained in the README.rst; in all other cases you may"
                           " want to report the issue.")

with open('version.txt', encoding='utf-8') as f:
    VERSION = f.read()

with open('README.rst', encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()

with open('CHANGES.rst', encoding='utf-8') as f:
    CHANGES = f.read()

extension_options = {
    'sources': ['./rapidjson.cpp'],
    'include_dirs': [rj_include_dir],
    'define_macros': [('PYTHON_RAPIDJSON_VERSION', VERSION)],
    'depends': ['./rapidjson_exact_version.txt'],
}

if os.path.exists('rapidjson_exact_version.txt'):
    with open('rapidjson_exact_version.txt', encoding='utf-8') as f:
        extension_options['define_macros'].append(
            ('RAPIDJSON_EXACT_VERSION', f.read().strip()))


cxx = sysconfig.get_config_var('CXX')
if cxx and 'g++' in cxx:
    # Avoid warning about invalid flag for C++
    for varname in ('CFLAGS', 'OPT'):
        value = sysconfig.get_config_var(varname)
        if value and '-Wstrict-prototypes' in value:
            value = value.replace('-Wstrict-prototypes', '')
            sysconfig.get_config_vars()[varname] = value

    # Add -pedantic, so we get a warning when using non-standard features, and
    # -Wno-long-long to pacify old gcc (or Apple's hybrids) that treat "long long" as an
    # error under C++ (see issue #69). C++11 is required since commit
    # https://github.com/Tencent/rapidjson/commit/9965ab37f6cfae3d58a0a6e34c76112866ace0b1
    extension_options['extra_compile_args'] = [
        '-pedantic', '-Wno-long-long', '-std=c++11']

    # Up to Python 3.7, some structures use "char*" instead of "const char*",
    # and ISO C++ forbids assigning string literal constants
    if sys.version_info < (3,7):
        extension_options['extra_compile_args'].append('-Wno-write-strings')


setup(
    name='python-rapidjson',
    version=VERSION,
    description='Python wrapper around rapidjson',
    long_description=LONG_DESCRIPTION + '\n\n' + CHANGES,
    long_description_content_type='text/x-rst',
    license='MIT License',
    keywords='json rapidjson',
    author='Ken Robbins',
    author_email='ken@kenrobbins.com',
    maintainer='Lele Gaifax',
    maintainer_email='lele@metapensiero.it',
    url='https://github.com/python-rapidjson/python-rapidjson',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: C++',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python',
    ],
    ext_modules=[Extension('rapidjson', **extension_options)],
    **other_setup_options
)



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Basic params tests
# :Author:    Ken Robbins <ken@kenrobbins.com>
# :License:   MIT License
# :Copyright: © 2015 Ken Robbins
# :Copyright: © 2015, 2016, 2017, 2018, 2019, 2020 Lele Gaifax
#

from calendar import timegm
from datetime import date, datetime, time, timezone, timedelta
import io
import math
import uuid

import pytest

import rapidjson as rj


def test_skipkeys():
    o = {True: False, -1: 1, 1.1: 1.1, (1,2): "foo", b"asdf": 1, None: None}

    with pytest.raises(TypeError):
        rj.dumps(o)

    with pytest.raises(TypeError):
        rj.dumps(o, skipkeys=False)

    assert rj.dumps(o, skipkeys=True) == '{}'
    assert rj.dumps(o, mapping_mode=rj.MM_SKIP_NON_STRING_KEYS) == '{}'


def test_coerce_keys():
    o = {True: False, -1: 1, 1.1: 1.1, (1,2): "foo", b"asdf": 1, None: None}
    expected = '{"True":false,"-1":1,"1.1":1.1,"(1, 2)":"foo","b\'asdf\'":1,"None":null}'
    assert rj.dumps(o, mapping_mode=rj.MM_COERCE_KEYS_TO_STRINGS) == expected
    assert rj.Encoder(mapping_mode=rj.MM_COERCE_KEYS_TO_STRINGS)(o) == expected


def test_skip_invalid_keys():
    o = {True: False, -1: 1, 1.1: 1.1, (1,2): "foo", b"asdf": 1, None: None}

    with pytest.raises(TypeError):
        rj.Encoder()(o)

    with pytest.raises(TypeError):
        rj.Encoder(skip_invalid_keys=False)(o)

    assert rj.Encoder(skip_invalid_keys=True)(o) == '{}'
    assert rj.Encoder(mapping_mode=rj.MM_SKIP_NON_STRING_KEYS)(o) == '{}'


def test_ensure_ascii(dumps):
    s = '\N{GREEK SMALL LETTER ALPHA}\N{GREEK CAPITAL LETTER OMEGA}'
    assert dumps(s) == '"\\u03B1\\u03A9"'
    assert dumps(s, ensure_ascii=True) == '"\\u03B1\\u03A9"'
    assert dumps(s, ensure_ascii=False) == '"%s"' % s


def test_allow_nan():
    f = [1.1, float("inf"), 2.2, float("nan"), 3.3, float("-inf"), 4.4]
    expected = '[1.1,Infinity,2.2,NaN,3.3,-Infinity,4.4]'
    assert rj.dumps(f) == expected
    assert rj.dumps(f, number_mode=rj.NM_NAN) == expected
    assert rj.dumps(f, allow_nan=True) == expected
    with pytest.raises(ValueError):
        rj.dumps(f, number_mode=None)
    with pytest.raises(ValueError):
        rj.dumps(f, allow_nan=False)

    s = "NaN"
    assert math.isnan(rj.loads(s))
    assert math.isnan(rj.loads(s, number_mode=rj.NM_NAN))
    assert math.isnan(rj.loads(s, allow_nan=True))
    with pytest.raises(ValueError):
        rj.loads(s, number_mode=rj.NM_NONE)
    with pytest.raises(ValueError):
        rj.loads(s, allow_nan=False)

    s = "Infinity"
    assert rj.loads(s) == float("inf")
    assert rj.loads(s, number_mode=rj.NM_NAN) == float("inf")
    assert rj.loads(s, allow_nan=True) == float("inf")
    with pytest.raises(ValueError):
        rj.loads(s, number_mode=rj.NM_NONE)
    with pytest.raises(ValueError):
        rj.loads(s, allow_nan=False)

    s = "-Infinity"
    assert rj.loads(s) == float("-inf")
    assert rj.loads(s, number_mode=rj.NM_NAN) == float("-inf")
    assert rj.loads(s, allow_nan=True) == float("-inf")
    with pytest.raises(ValueError):
        rj.loads(s, number_mode=rj.NM_NONE)
    with pytest.raises(ValueError):
        rj.loads(s, allow_nan=False)


def test_native(dumps, loads):
    f = [-1, 1, 1.1, -2.2]
    expected = '[-1,1,1.1,-2.2]'
    assert dumps(f, number_mode=rj.NM_NATIVE) == expected
    assert dumps(f) == expected
    assert loads(expected, number_mode=rj.NM_NATIVE) == f
    assert loads(expected) == f

    trailing_comma = '[-1,1,1.1,-2.2,]'
    pytest.raises(ValueError, loads, trailing_comma, number_mode=rj.NM_NATIVE)
    expected = [-1,1,1.1,-2.2]
    assert loads(trailing_comma, number_mode=rj.NM_NATIVE,
                 parse_mode=rj.PM_TRAILING_COMMAS) == expected

    comments = '[-1,1,/*1.1,*/-2.2,]'
    pytest.raises(ValueError, loads, comments, number_mode=rj.NM_NATIVE)
    pytest.raises(ValueError, loads, comments,
                  number_mode=rj.NM_NATIVE, parse_mode=rj.PM_TRAILING_COMMAS)
    expected = [-1,1,-2.2]
    assert loads(comments, number_mode=rj.NM_NATIVE,
                 parse_mode=rj.PM_COMMENTS | rj.PM_TRAILING_COMMAS) == expected


def test_parse_mode(dumps, loads):
    trailing_comma = '[-1,1,1.1,-2.2,]'
    expected = [-1,1,1.1,-2.2]
    pytest.raises(ValueError, rj.loads, trailing_comma)
    pytest.raises(ValueError, rj.loads, trailing_comma, parse_mode=rj.PM_COMMENTS)
    assert loads(trailing_comma, parse_mode=rj.PM_TRAILING_COMMAS) == expected

    comments = ('[true,  // boolean\n'
                ' false, // idem\n'
                ' // 1.0, // ignored\n'
                ' /*\n'
                '  * ignored\n'
                ' 1,     // integer\n'
                '  */'
                ' "this is the end"]')
    expected = [True, False, "this is the end"]
    pytest.raises(ValueError, loads, comments)
    pytest.raises(ValueError, loads, comments, parse_mode=rj.PM_TRAILING_COMMAS)
    assert loads(comments, parse_mode=rj.PM_COMMENTS) == expected

    c_and_tc = ('[true,  // boolean\n'
                ' false, // idem\n'
                ' // 1.0, // ignored\n'
                ' /*\n'
                '  * ignored\n'
                ' 1,     // integer\n'
                '  */'
                ' {"one": 1 /*, "two": 2 */, "three": 3,},\n'
                ' "this is the end",]'
    )
    expected = [True, False, {"one": 1, "three": 3}, "this is the end"]
    pytest.raises(ValueError, loads, c_and_tc)
    pytest.raises(ValueError, loads, c_and_tc, parse_mode=rj.PM_TRAILING_COMMAS)
    pytest.raises(ValueError, loads, c_and_tc, parse_mode=rj.PM_COMMENTS)
    assert loads(c_and_tc, parse_mode=rj.PM_COMMENTS | rj.PM_TRAILING_COMMAS) == expected


def test_indent(dumps):
    o = {"a": 1, "z": 2, "b": 3}
    expected1 = '{\n    "a": 1,\n    "z": 2,\n    "b": 3\n}'
    expected2 = '{\n    "a": 1,\n    "b": 3,\n    "z": 2\n}'
    expected3 = '{\n    "b": 3,\n    "a": 1,\n    "z": 2\n}'
    expected4 = '{\n    "b": 3,\n    "z": 2,\n    "a": 1\n}'
    expected5 = '{\n    "z": 2,\n    "a": 1,\n    "b": 3\n}'
    expected6 = '{\n    "z": 2,\n    "b": 3,\n    "a": 1\n}'
    expected = (
        expected1,
        expected2,
        expected3,
        expected4,
        expected5,
        expected6)

    assert dumps(o, indent=4) in expected
    assert dumps(o, indent="    ") in expected
    assert dumps(o, indent="\t") in (e.replace("    ", "\t") for e in expected)

    with pytest.raises(TypeError):
        dumps(o, indent="\n\t")

    with pytest.raises(TypeError):
        dumps(o, indent="foo")

    with pytest.raises(TypeError):
        dumps(o, indent=-1)


def test_write_mode(dumps):
    o = {"a": 1, "b": [2, 3, 4]}
    expected_compact = ('{"a":1,"b":[2,3,4]}', '{"b":[2,3,4],"a":1}')
    expected_pretty = ('{\n    "a": 1,\n    "b": [2, 3, 4]\n}',
                       '{\n    "b": [2, 3, 4],\n    "a": 1\n}')

    assert dumps(o, write_mode=rj.WM_COMPACT) in expected_compact
    assert dumps(o, indent=0, write_mode=rj.WM_COMPACT) in expected_compact
    assert dumps(o, write_mode=rj.WM_SINGLE_LINE_ARRAY) in expected_pretty
    assert dumps(o, write_mode=rj.WM_PRETTY|rj.WM_SINGLE_LINE_ARRAY) in expected_pretty

    with pytest.raises(ValueError):
        dumps(o, write_mode=4)


def test_sort_keys(dumps):
    o = {"a": 1, "z": 2, "b": 3}
    expected0 = '{\n"a": 1,\n"b": 3,\n"z": 2\n}'
    expected1 = '{"a":1,"b":3,"z":2}'
    expected2 = '{\n    "a": 1,\n    "b": 3,\n    "z": 2\n}'

    assert dumps(o, sort_keys=True) == expected1
    assert dumps(o, sort_keys=True, indent=4) == expected2
    assert dumps(o, sort_keys=True, indent=0) == expected0
    assert dumps(o, mapping_mode=rj.MM_SORT_KEYS, indent=0) == expected0

    o = {'a0': 'a0', 'a': 'a', 'a1': 'a1', 'a\x00b': 'a\x00b'}
    assert sorted(o.keys()) == ['a', 'a\x00b', 'a0', 'a1']
    assert dumps(o, sort_keys=True) == '{"a":"a","a\\u0000b":"a\\u0000b","a0":"a0","a1":"a1"}'


def test_default():
    class Bar:
        pass

    class Foo:
        def __init__(self):
            self.foo = "bar"

    def default(obj):
        if isinstance(obj, Foo):
            return {"foo": obj.foo}

        raise TypeError("default error")

    o = {"asdf": Foo()}
    assert rj.dumps(o, default=default) == '{"asdf":{"foo":"bar"}}'

    o = {"asdf": Foo(), "qwer": Bar()}
    with pytest.raises(TypeError):
        rj.dumps(o, default=default)

    with pytest.raises(TypeError):
        rj.dumps(o)


def test_decimal(dumps, loads):
    import math
    from decimal import Decimal

    dstr = "2.7182818284590452353602874713527"
    d = Decimal(dstr)

    with pytest.raises(TypeError):
        dumps(d)

    assert dumps(float(dstr)) == str(math.e)
    assert dumps(d, number_mode=rj.NM_DECIMAL) == dstr
    assert dumps({"foo": d}, number_mode=rj.NM_DECIMAL) == '{"foo":%s}' % dstr

    assert loads(dumps(d, number_mode=rj.NM_DECIMAL), number_mode=rj.NM_DECIMAL) == d

    assert loads(dumps(d, number_mode=rj.NM_DECIMAL)) == float(dstr)


def test_datetime_mode_dumps(dumps):
    import pytz

    d = datetime(2018, 1, 2, 11, 56, 19, 854440)
    dstr = d.isoformat()

    with pytest.raises(TypeError):
        dumps(d)

    with pytest.raises(TypeError):
        dumps(d, datetime_mode=rj.DM_NONE)

    assert dumps(d, datetime_mode=rj.DM_ISO8601) == '"%s"' % dstr
    assert dumps(d, datetime_mode=(rj.DM_ISO8601 | rj.DM_IGNORE_TZ)) == '"%s"' % dstr

    d = utcd = d.replace(tzinfo=pytz.utc)
    dstr = utcstr = d.isoformat()

    assert dumps(d, datetime_mode=rj.DM_ISO8601) == '"%s"' % dstr
    assert dumps(d, datetime_mode=(rj.DM_ISO8601 | rj.DM_IGNORE_TZ)) == '"%s"' % dstr[:-6]

    d = d.astimezone(pytz.timezone('Pacific/Chatham'))
    dstr = d.isoformat()

    assert dumps(d, datetime_mode=rj.DM_ISO8601) == '"%s"' % dstr
    assert dumps(d, datetime_mode=(rj.DM_ISO8601 | rj.DM_IGNORE_TZ)) == '"%s"' % dstr[:-6]

    d = d.astimezone(pytz.timezone('Asia/Kathmandu'))
    dstr = d.isoformat()

    assert dumps(d, datetime_mode=rj.DM_ISO8601) == '"%s"' % dstr
    assert dumps(d, datetime_mode=(rj.DM_ISO8601 | rj.DM_IGNORE_TZ)) == '"%s"' % dstr[:-6]

    d = d.astimezone(pytz.timezone('America/New_York'))
    dstr = d.isoformat()

    assert dumps(d, datetime_mode=rj.DM_ISO8601) == '"%s"' % dstr
    assert dumps(d, datetime_mode=(rj.DM_ISO8601 | rj.DM_IGNORE_TZ)) == '"%s"' % dstr[:-6]
    assert dumps(d, datetime_mode=(rj.DM_ISO8601 | rj.DM_SHIFT_TO_UTC)) == '"%s"' % utcstr

    assert dumps(d, datetime_mode=rj.DM_UNIX_TIME) == str(d.timestamp())

    assert dumps(
        d, datetime_mode=rj.DM_UNIX_TIME | rj.DM_SHIFT_TO_UTC) == str(utcd.timestamp())

    assert dumps(
        d, datetime_mode= rj.DM_UNIX_TIME | rj.DM_ONLY_SECONDS
    ) == str(d.timestamp()).split('.')[0]

    d = datetime(2018, 1, 2, 12, 56, 19, 854440)

    assert dumps(
        d, datetime_mode=rj.DM_ISO8601 | rj.DM_NAIVE_IS_UTC
    ) == '"%s+00:00"' % d.isoformat()

    assert dumps(
        d, datetime_mode=rj.DM_UNIX_TIME | rj.DM_NAIVE_IS_UTC
    ) == ('%d.%06d' % (timegm(d.timetuple()), d.microsecond)).rstrip('0')

    assert dumps(
        d, datetime_mode=(rj.DM_UNIX_TIME
                          | rj.DM_NAIVE_IS_UTC
                          | rj.DM_ONLY_SECONDS)
    ) == str(timegm(d.timetuple()))

    # This value caused a truncation problem when we were using
    # Writer.SetMaxDecimalPlaces(6) to emit the floating point number
    d = datetime.fromtimestamp(1514893636.276703)

    assert dumps(d, datetime_mode=rj.DM_UNIX_TIME) == str(d.timestamp())


def test_datetime_mode_loads(dumps, loads):
    import pytz

    utc = datetime.now(pytz.utc)
    utcstr = utc.isoformat()

    jsond = dumps(utc, datetime_mode=rj.DM_ISO8601)

    assert jsond == '"%s"' % utcstr
    assert loads(jsond, datetime_mode=rj.DM_ISO8601) == utc

    local = utc.astimezone(pytz.timezone('Europe/Rome'))
    locstr = local.isoformat()

    jsond = dumps(local, datetime_mode=rj.DM_ISO8601)

    assert jsond == '"%s"' % locstr
    assert loads(jsond) == locstr
    assert loads(jsond, datetime_mode=rj.DM_ISO8601) == local

    load_as_utc = loads(jsond, datetime_mode=(rj.DM_ISO8601 | rj.DM_SHIFT_TO_UTC))

    assert load_as_utc == utc
    assert not load_as_utc.utcoffset()

    load_as_naive = loads(jsond, datetime_mode=(rj.DM_ISO8601 | rj.DM_IGNORE_TZ))

    assert load_as_naive == local.replace(tzinfo=None)


@pytest.mark.parametrize(
    'value', [date.today(), datetime.now(), time(10,20,30)])
def test_datetime_values(value, dumps, loads):
    with pytest.raises(TypeError):
        rj.dumps(value)

    dumped = dumps(value, datetime_mode=rj.DM_ISO8601)
    loaded = loads(dumped, datetime_mode=rj.DM_ISO8601)
    assert loaded == value


@pytest.mark.parametrize(
    'value, expected', [
        ('1999-01-03T10:20:30.1',
         datetime(1999, 1, 3, 10, 20, 30, 100000)),
        ('2000-02-03T10:20:30.12',
         datetime(2000, 2, 3, 10, 20, 30, 120000)),
        ('2020-03-03T10:20:30.123',
         datetime(2020, 3, 3, 10, 20, 30, 123000)),
        ('2021-04-03T10:20:30.1234',
         datetime(2021, 4, 3, 10, 20, 30, 123400)),
        ('1999-08-03T10:20:30.12345',
         datetime(1999, 8, 3, 10, 20, 30, 123450)),
        ('1999-09-03T10:20:30.123456',
         datetime(1999, 9, 3, 10, 20, 30, 123456)),
        ('1999-10-03T10:20:30.1234567',
         datetime(1999, 10, 3, 10, 20, 30, 123456)),
        ('1999-11-03T10:20:30.12345678',
         datetime(1999, 11, 3, 10, 20, 30, 123456)),
        ('1999-12-03T10:20:30.123456789',
         datetime(1999, 12, 3, 10, 20, 30, 123456)),
        ('1999-12-03T10:20:30.123456789Z',
         datetime(1999, 12, 3, 10, 20, 30, 123456, tzinfo=timezone.utc)),
        ('1999-12-03T10:20:30.123456789-02:30',
         datetime(1999, 12, 3, 10, 20, 30, 123456,
                  tzinfo=timezone(timedelta(days=-1, seconds=77400)))),
        ('1999-12-03T10:20:30.123456789+02:30',
         datetime(1999, 12, 3, 10, 20, 30, 123456,
                  tzinfo=timezone(timedelta(seconds=9000)))),
    ]
)
def test_datetime_fractional_values(value, expected, dumps, loads):
    dumped = dumps(value, datetime_mode=rj.DM_ISO8601)
    loaded = loads(dumped, datetime_mode=rj.DM_ISO8601)
    assert loaded == expected


def test_uuid_mode(dumps, loads):
    assert rj.UM_NONE == 0
    assert rj.UM_CANONICAL == 1
    assert rj.UM_HEX == 2

    value = uuid.uuid1()
    with pytest.raises(TypeError):
        dumps(value)

    with pytest.raises(ValueError):
        dumps(value, uuid_mode=42)

    with pytest.raises(ValueError):
        loads('""', uuid_mode=42)

    dumped = dumps(value, uuid_mode=rj.UM_CANONICAL)
    loaded = loads(dumped, uuid_mode=rj.UM_CANONICAL)
    assert loaded == value

    # When loading, hex mode implies canonical format
    loaded = loads(dumped, uuid_mode=rj.UM_HEX)
    assert loaded == value

    dumped = dumps(value, uuid_mode=rj.UM_HEX)
    loaded = loads(dumped, uuid_mode=rj.UM_HEX)
    assert loaded == value


def test_uuid_and_datetime_mode_together(dumps, loads):
    value = [date.today(), uuid.uuid1()]
    dumped = dumps(value,
                   datetime_mode=rj.DM_ISO8601,
                   uuid_mode=rj.UM_CANONICAL)
    loaded = loads(dumped,
                   datetime_mode=rj.DM_ISO8601,
                   uuid_mode=rj.UM_CANONICAL)
    assert loaded == value


@pytest.mark.parametrize(
    'value,cls', [
        ('x999-02-03', str),
        ('1999 02 03', str),
        ('x0:02:20', str),
        ('20.02:20', str),
        ('x999-02-03T10:20:30', str),
        ('1999-02-03t10:20:30', str),

        ('0000-01-01', str),
        ('0001-99-99', str),
        ('0001-01-32', str),
        ('0001-02-29', str),

        ('24:02:20', str),
        ('23:62:20', str),
        ('23:02:62', str),
        ('20:02:20.123-25:00', str),
        ('20:02:20.123-05:61', str),

        ('1968-02-29', date),
        ('1999-02-03', date),

        ('20:02:20', time),
        ('20:02:20Z', time),
        ('20:02:20.123', time),
        ('20:02:20.123Z', time),
        ('20:02:20-05:00', time),
        ('20:02:20.123456', time),
        ('20:02:20.123456Z', time),
        ('20:02:20.123-05:00', time),
        ('20:02:20.123456-05:00', time),

        ('1999-02-03T10:20:30', datetime),
        ('1999-02-03T10:20:30Z', datetime),
        ('1999-02-03T10:20:30.1', datetime),
        ('1999-02-03T10:20:30.12', datetime),
        ('1999-02-03T10:20:30.123', datetime),
        ('1999-02-03T10:20:30.123Z', datetime),
        ('1999-02-03T10:20:30.1234', datetime),
        ('1999-02-03T10:20:30.12345', datetime),
        ('1999-02-03T10:20:30.12345Z', datetime),
        ('1999-02-03T10:20:30.12345+05:00', datetime),
        ('1999-02-03T10:20:30-05:00', datetime),
        ('1999-02-03T10:20:30.123456', datetime),
        ('1999-02-03T10:20:30.123456Z', datetime),
        ('1999-02-03T10:20:30.123-05:00', datetime),
        ('1999-02-03T10:20:30.123456-05:00', datetime),
        ('1999-02-03T10:20:30.123456-05:00', datetime),
        ('1999-02-03T10:20:30.1234567', datetime),
        ('1999-02-03T10:20:30.12345678', datetime),
        ('1999-02-03T10:20:30.123456789', datetime),
        ('1999-02-03T10:20:30.1234567Z', datetime),
        ('1999-02-03T10:20:30.12345678Z', datetime),
        ('1999-02-03T10:20:30.123456789Z', datetime),
        ('1999-02-03T10:20:30.1234567-20:10', datetime),
        ('1999-02-03T10:20:30.12345678+20:11', datetime),
        ('1999-02-03T10:20:30.123456789Z', datetime),
        ('1999-02-03T10:20:30.123456789-01:01', datetime),
        ('1999-02-03T10:20:30.123456789+00:00', datetime),
        ('1999-02-03T10:20:30.', str),
        ('1999-02-03T10:20:30.1A', str),
        ('1999-02-03T10:20:30. ', str),
        ('1999-02-03T10:20:30.123456789A', str),
        ('1999-02-03T10:20:30.123456789 ', str),
        ('1999-02-03T10:20:30.12345+00:00Z ', str),
        ('1999-02-03T10:20:30.1A34X6+01:00Z', str),
    ])
def test_datetime_iso8601(value, cls, loads):
    result = loads('"%s"' % value, datetime_mode=rj.DM_ISO8601)
    assert isinstance(result, cls)


@pytest.mark.parametrize(
    'value,cls', [
        ('7a683da49aa011e5972e3085a99ccac7', str),
        ('7a683da4 9aa0-11e5-972e-3085a99ccac7', str),
        ('za683da4-9aa0-11e5-972e-3085a99ccac7', str),

        ('7a683da4-9aa0-11e5-972e-3085a99ccac7', uuid.UUID),
    ])
def test_uuid_canonical(value, cls, loads):
    result = loads('"%s"' % value, uuid_mode=rj.UM_CANONICAL)
    assert isinstance(result, cls), type(result)


@pytest.mark.parametrize(
    'value,cls', [
        ('za683da49aa011e5972e3085a99ccac7', str),

        ('7a683da49aa011e5972e3085a99ccac7', uuid.UUID),
        ('7a683da4-9aa0-11e5-972e-3085a99ccac7', uuid.UUID),
    ])
def test_uuid_hex(value, cls, loads):
    result = loads('"%s"' % value, uuid_mode=rj.UM_HEX)
    assert isinstance(result, cls), type(result)


def test_object_hook():
    class Foo:
        def __init__(self, foo):
            self.foo = foo

    def hook(d):
        if 'foo' in d:
            return Foo(d['foo'])

        return d

    def default(obj):
        return {'foo': obj.foo}

    res = rj.loads('{"foo": 1}', object_hook=hook)
    assert isinstance(res, Foo)
    assert res.foo == 1

    assert rj.dumps(rj.loads('{"foo": 1}', object_hook=hook), default=default) == '{"foo":1}'
    res = rj.loads(rj.dumps(Foo(foo="bar"), default=default), object_hook=hook)
    assert isinstance(res, Foo)
    assert res.foo == "bar"


@pytest.mark.parametrize(
    'posargs,kwargs', (
        ( (), {} ),
        ( (None,), {} ),
        ( (True,), {} ),
        ( ('{}',), { 'this_keyword_arg_shall_never_exist': True } ),
        ( ('[]',), { 'object_hook': True } ),
        ( ('[]',), { 'datetime_mode': 'no' } ),
        ( ('[]',), { 'datetime_mode': 1.0 } ),
        ( ('[]',), { 'datetime_mode': -100 } ),
        ( ('[]',), { 'datetime_mode': rj.DM_UNIX_TIME + 1 } ),
        ( ('[]',), { 'datetime_mode': rj.DM_UNIX_TIME } ),
        ( ('[]',), { 'datetime_mode': rj.DM_SHIFT_TO_UTC } ),
        ( ('[]',), { 'uuid_mode': 'no' } ),
        ( ('[]',), { 'uuid_mode': 1.0 } ),
        ( ('[]',), { 'uuid_mode': -100 } ),
        ( ('[]',), { 'uuid_mode': 100 } ),
        ( ('[]',), { 'parse_mode': 'no' } ),
        ( ('[]',), { 'parse_mode': 1.0 } ),
        ( ('[]',), { 'uuid_mode': -100 } ),
    ))
def test_invalid_loads_params(posargs, kwargs, loads):
    try:
        loads(*posargs, **kwargs)
    except (TypeError, ValueError):
        pass
    else:
        assert False, "Expected either a TypeError or a ValueError"


@pytest.mark.parametrize(
    'posargs,kwargs', (
        ( (), {} ),
        ( ([],), { 'this_keyword_arg_shall_never_exist': True } ),
        ( ([],), { 'default': True } ),
        ( ([],), { 'indent': -1 }),
        ( ([],), { 'datetime_mode': 'no' } ),
        ( ([],), { 'datetime_mode': 1.0 } ),
        ( ([],), { 'datetime_mode': -100 } ),
        ( ([],), { 'datetime_mode': rj.DM_UNIX_TIME + 1 } ),
        ( ([],), { 'datetime_mode': rj.DM_SHIFT_TO_UTC } ),
        ( ([],), { 'uuid_mode': 'no' } ),
        ( ([],), { 'uuid_mode': 1.0 } ),
        ( ([],), { 'uuid_mode': -100 } ),
        ( ([],), { 'uuid_mode': 100 } ),
        ( ([],), { 'iterable_mode': 'no'} ),
        ( ([],), { 'iterable_mode': 1.0} ),
        ( ([],), { 'iterable_mode': -100} ),
        ( ([],), { 'iterable_mode': 100} ),
        ( ([],), { 'mapping_mode': 'no'} ),
        ( ([],), { 'mapping_mode': 1.0} ),
        ( ([],), { 'mapping_mode': -100} ),
        ( ([],), { 'mapping_mode': 100} ),
    ))
def test_invalid_dumps_params(posargs, kwargs, dumps):
    try:
        dumps(*posargs, **kwargs)
    except (TypeError, ValueError):
        pass
    else:
        assert False, "Expected either a TypeError or a ValueError"


def test_explicit_defaults_loads():
    assert rj.loads(
        string='"foo"',
        object_hook=None,
        number_mode=None,
        datetime_mode=None,
        uuid_mode=None,
        parse_mode=None,
        allow_nan=True,
    ) == "foo"


def test_explicit_defaults_load():
    assert rj.load(
        stream=io.StringIO('"foo"'),
        object_hook=None,
        number_mode=None,
        datetime_mode=None,
        uuid_mode=None,
        parse_mode=None,
        chunk_size=None,
        allow_nan=True,
    ) == "foo"


def test_explicit_defaults_decoder():
    assert rj.Decoder(
        number_mode=None,
        datetime_mode=None,
        uuid_mode=None,
        parse_mode=None)('"foo"') == "foo"


def test_explicit_defaults_dumps():
    assert rj.dumps(
        obj='foo',
        skipkeys=False,
        ensure_ascii=True,
        indent=None,
        default=None,
        sort_keys=False,
        number_mode=None,
        datetime_mode=None,
        uuid_mode=None,
        allow_nan=True,
    ) == '"foo"'


def test_explicit_defaults_dump():
    stream = io.StringIO()
    assert rj.dump(
        obj='foo',
        stream=stream,
        skipkeys=False,
        ensure_ascii=True,
        indent=None,
        default=None,
        sort_keys=False,
        number_mode=None,
        datetime_mode=None,
        uuid_mode=None,
        allow_nan=True,
        chunk_size=None) is None
    assert stream.getvalue() == '"foo"'


def test_explicit_defaults_encoder():
    assert rj.Encoder(
        skip_invalid_keys=False,
        ensure_ascii=True,
        indent=None,
        sort_keys=False,
        number_mode=None,
        datetime_mode=None,
        uuid_mode=None)({'foo': 'bar'}) == '{"foo":"bar"}'


def test_encoder_call():
    o = {'foo': 'bar'}

    result = rj.Encoder()(o)
    assert result == '{"foo":"bar"}'
    assert rj.Encoder()(o, None) == result

    stream = io.StringIO()
    assert rj.Encoder()(o, stream) is None
    assert stream.getvalue() == result

    stream = io.StringIO()
    assert rj.Encoder()(o, stream=stream) is None
    assert stream.getvalue() == result


def test_decoder_attrs():
    d = rj.Decoder(
        number_mode=rj.NM_NAN,
        datetime_mode=rj.DM_ISO8601,
        uuid_mode=rj.UM_CANONICAL,
        parse_mode=rj.PM_COMMENTS)
    assert d.number_mode == rj.NM_NAN
    assert d.datetime_mode == rj.DM_ISO8601
    assert d.uuid_mode == rj.UM_CANONICAL
    assert d.parse_mode == rj.PM_COMMENTS


def test_encoder_attrs():
    e = rj.Encoder(
        skip_invalid_keys=True,
        ensure_ascii=True,
        indent='\n',
        sort_keys=True,
        number_mode=rj.NM_NAN,
        datetime_mode=rj.DM_ISO8601,
        uuid_mode=rj.UM_CANONICAL,
        bytes_mode=rj.BM_UTF8)
    assert e.skip_invalid_keys
    assert e.ensure_ascii
    assert e.indent_char == '\n'
    assert e.indent_count == 1
    assert e.write_mode == rj.WM_PRETTY
    assert e.sort_keys
    assert e.number_mode == rj.NM_NAN
    assert e.datetime_mode == rj.DM_ISO8601
    assert e.uuid_mode == rj.UM_CANONICAL
    assert e.bytes_mode == rj.BM_UTF8

    e = rj.Encoder(mapping_mode=rj.MM_SKIP_NON_STRING_KEYS|rj.MM_SORT_KEYS)
    assert e.skip_invalid_keys
    assert e.sort_keys



# -*- coding: utf-8 -*-
# :Project:   python-rapidjson -- Refs leaks tests
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   MIT License
# :Copyright: © 2017, 2018, 2020, 2022 Lele Gaifax
#

# NB: this is a simplicistic test that uses sys.gettotalrefcount(), available
# when the interpreter is built --with-pydebug, that tries to assert that
# repeated calls to dumps() and loads() does not leak object references.
# Since it's not an exact science, it should be taken with a grain of salt.

import datetime
import io
import sys
import uuid

import pytest

import rapidjson as rj


class Foo:
    def __init__(self, foo):
        self.foo = foo


def hook(d):
    if 'foo' in d:
        return Foo(d['foo'])
    return d


def default(obj):
    return {'foo': obj.foo}


# Empirical tolerance used to test the refcount growing
THRESHOLD = 7


plain_string = 'plain-string'
bigint = 123456789
pi = 3.1415926
right_now = datetime.datetime.now()
date = right_now.date()
time = right_now.time()
nil = uuid.UUID(int=0)
foo = Foo('foo')
array = [dict(index=j, array=[dict(index=i,
                                   plain_string=plain_string,
                                   bigint=bigint,
                                   pi=pi,
                                   right_now=right_now,
                                   date=date,
                                   time=time,
                                   nil=nil,
                                   foos=[foo, foo, foo, foo, foo])
                              for i in range(10)])
         for j in range(10)]

NO_OPTION = {}
DATETIMES = {'datetime_mode': rj.DM_ISO8601}
UUIDS = {'uuid_mode': rj.UM_CANONICAL}
FOOS_DUMP = {'default': default}
FOOS_LOAD = {'object_hook': hook}
ARRAY_DUMP = {'datetime_mode': rj.DM_ISO8601,
              'uuid_mode': rj.UM_CANONICAL,
              'default': default}
ARRAY_LOAD = {'datetime_mode': rj.DM_ISO8601,
              'uuid_mode': rj.UM_CANONICAL,
              'object_hook': hook}

@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
@pytest.mark.parametrize('value,dumps_options,loads_options', [
    ( plain_string, NO_OPTION, NO_OPTION ),
    ( bigint, NO_OPTION, NO_OPTION ),
    ( pi, NO_OPTION, NO_OPTION ),
    ( rj.RawJSON(' "foo" '), NO_OPTION, NO_OPTION),
    ( right_now, DATETIMES, DATETIMES ),
    ( date, DATETIMES, DATETIMES ),
    ( time, DATETIMES, DATETIMES ),
    ( nil, UUIDS, UUIDS ),
    ( foo, FOOS_DUMP, FOOS_LOAD ),
    ( array, ARRAY_DUMP, ARRAY_LOAD ),
])
def test_leaks(value, dumps_options, loads_options):
    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        asjson = rj.dumps(value, **dumps_options)
        aspython = rj.loads(asjson, **loads_options)
        del asjson
        del aspython
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD

    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        stream = io.BytesIO()
        none = rj.dump(value, stream, **dumps_options)
        stream.seek(0)
        aspython = rj.load(stream, **loads_options)
        del none
        del aspython
        del stream
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
def test_dump_iterator_leaks():
    rc0 = sys.gettotalrefcount()
    value = iter(range(10000))
    asjson = rj.dumps(value)
    del asjson
    del value
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
def test_decoder_call_leaks():
    class MyDecoder(rj.Decoder):
        def start_object(self):
            return {}
        def end_object(self, obj):
            return obj
        def end_array(self, array):
            return array
        def string(self, string):
            return string

    decoder = MyDecoder()
    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        value = decoder('["foo", {"foo": "bar"}]')
        del value
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


@pytest.mark.parametrize('value', ['Foo', rj.RawJSON('Foo')])
@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
def test_encoder_call_leaks(value):
    class MyEncoder(rj.Encoder):
        def __init__(self, value):
            self.value = value

        def default(self, obj):
            return self.value

    class Foo:
        pass

    encoder = MyEncoder(value)
    foo = Foo()
    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        value = encoder(foo)
        del value
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
def test_rawjson_constructor():
    raw_json = rj.RawJSON('["foo", "bar"]')
    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        value = '"foobar"'
        raw_json.__init__(value)
        del value
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
def test_rawjson_new():
    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        raw_json = rj.RawJSON('["foo", "bar"]')
        del raw_json
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


@pytest.mark.skipif(not hasattr(sys, 'gettotalrefcount'), reason='Non-debug Python')
def test_invalid_json_load_leaks():
    # See issue #148
    value = '{"a":{"b":}}'
    rc0 = sys.gettotalrefcount()
    for _ in range(1000):
        try:
            rj.loads(value)
        except rj.JSONDecodeError:
            pass
    del _
    rc1 = sys.gettotalrefcount()
    assert (rc1 - rc0) < THRESHOLD


def test_endarray_leak():
    # See issue #160
    value = '{"v": [1, 2]}'

    class Decoder1(rj.Decoder):
        pass

    class Decoder2(rj.Decoder):
        def end_array(self, seq):
            return list(seq)

    j1 = rj.loads(value)
    assert sys.getrefcount(j1['v']) == 3

    j2 = Decoder1()(value)
    assert sys.getrefcount(j2['v']) == 3

    j3 = Decoder2()(value)
    assert sys.getrefcount(j3['v']) == 3
